{ csv2bfg - Converting a CSV file to a BFG input file

  This file is part of
  BFG - Buchhaltung fr Geeks - bookkeeping for geeks.

  Copyright (C) 2004 Peter Gerwinski <peter@gerwinski.de>

  BFG is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published
  by the Free Software Foundation, version 2.

  BFG is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with BFG; see the file COPYING. If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
  MA 02111-1307, USA. }

program csv2bfg;

uses
  GPC;

type
  PStringRec = ^TStringRec;

  TStringRec = record
    S: PString;
    Next: PStringRec
  end;

var
  InFileName, OutFileName: TString = '-';
  IncludeFiles: TString = '';
  InFile, OutFile: Text;
  ColumnName: PStringRec = nil;
  LineNo: Integer = 0;

procedure Init;
const
  LongOptions: array [1..2] of OptionType =
    (('output', RequiredArgument, nil, 'o'),
     ('include', RequiredArgument, nil, 'i'));
var
  LongIndex, i: Integer;
  Ch: Char;
begin
  LongIndex := -1;
  repeat
    Ch := GetOptLong ('', LongOptions, LongIndex, False);
    case Ch of
      'o': OutFileName := OptionArgument;
      'i': begin
             if IncludeFiles <> '' then
               IncludeFiles := IncludeFiles + NewLine;
             IncludeFiles := IncludeFiles + 'include "' + OptionArgument + '"'
           end;
    end
  until Ch = EndOfOptions;
  if FirstNonOption <= ParamCount then
    InFileName := ParamStr (FirstNonOption);
  Assign (InFile, InFileName);
  Reset (InFile);
  Assign (OutFile, OutFileName);
  Rewrite (OutFile)
end;

procedure Done;
begin
  Close (InFile);
  Close (OutFile)
end;

procedure Error (const Message: String);
begin
  if InFileName = '-' then
    Write (StdErr, 'line ', LineNo)
  else
    Write (StdErr, InFileName, ':', LineNo);
  WriteLn (StdErr, ': ', Message)
end;

procedure ConvertUmlauts (var Line: String);
var
  i: Integer;
begin
  for i := 1 to Length (Line) do
    case Line[i] of
      '': Line[i] := '';
      '': Line[i] := '';
      '': Line[i] := '';
      '': Line[i] := '';
      '': Line[i] := '';
      '': Line[i] := '';
      '': Line[i] := '';
    end
end;

procedure ExtractString (var Line, S: String);
var
  p: Integer;
begin
  if (Line = '') or (Line[1] <> '"') then
    begin
      Error ('`"'' expected');
      Line := '';
      S := '';
      Exit
    end;
  Delete (Line, 1, 1);
  p := Pos ('"', Line);
  if p <= 0 then
    begin
      Error ('second `"'' expected');
      S := Line;
      Line := '';
      Exit
    end;
  S := Copy (Line, 1, p - 1);
  Delete (Line, 1, p);
  if Line <> '' then
    if Line[1] = ';' then
      Delete (Line, 1, 1)
    else
      Error ('missing separator or garbage after string value')
end;

procedure AnalyseFirstLine;
var
  Line, S: TString;
  LastColumnName, NewColumnName: PStringRec;
begin
  LastColumnName:= ColumnName;
  while (LastColumnName <> nil) and (LastColumnName^.Next <> nil) do
    LastColumnName := LastColumnName^.Next;
  Inc (LineNo);
  ReadLn (InFile, Line);
  ConvertUmlauts (Line);
  while Line <> '' do
    begin
      ExtractString (Line, S);
      New (NewColumnName);
      NewColumnName^.S := NewString (S);
      NewColumnName^.Next := nil;
      if LastColumnName = nil then
        ColumnName := NewColumnName
      else
        LastColumnName^.Next := NewColumnName;
      LastColumnName := NewColumnName
    end
end;

function GetColumnString (const Line, Name: String): TString;
var
  p0, p, l: Integer;
  C: PStringRec;
  S: TString;
begin
  p0 := 1;
  p := 1;
  l := Length (Line);
  C := ColumnName;
  while C <> nil do
    begin
      while (p <= l) and (Line[p] <> ';') do
        begin
          if Line[p] = '"' then
            repeat
              Inc (p)
            until (p > l) or (Line[p] = '"');
          Inc (p)
        end;
      if (C^.S <> nil) and (C^.S^ = Name) then
        begin
          S := Copy (Line, p0, p - p0);
          if (S <> '') and (S[1] = '"') and (S[Length (S)] = '"') then
            S := Copy (S, 2, Length (S) - 2);
          GetColumnString := S;
          Exit
        end;
      C := C^.Next;
      p0 := p + 1;
      Inc (p)
    end;
  GetColumnString := ''
end;

function GetColumnReal (const Line, Name: String) = x: Real;
var
  S: TString;
  p: Integer;
begin
  S:= GetColumnString (Line, Name);
  p := Pos (',', S);
  if p > 0 then
    S[p] := '.';
  ReadStr (S, x)
end;

procedure ProcessFile;
var
  Line, Date, PrevDate, Command, VAT,
    Account, AntiAccount, Comment, TempStr: TString;
  Amount: Real;
  l: Integer;
begin
  Write (OutFile, '# generated by ', ParamStr (0));
  if InFileName <> '-' then
    Write (OutFile, ' from ', InFileName);
  WriteLn (OutFile);
  if IncludeFiles <> '' then
    begin
      WriteLn (OutFile);
      WriteLn (OutFile, IncludeFiles)
    end;
  PrevDate := '';
  while not EOF (InFile) do
    begin
      Inc (LineNo);
      ReadLn (InFile, Line);
      ConvertUmlauts (Line);
      Date := GetColumnString (Line, 'Belegdatum');
      l := Length (Date);
      Insert ('.', Date, l - 3);
      Insert ('.', Date, l - 5);
      if Date <> PrevDate then
        begin
          WriteLn (OutFile);
          WriteLn (OutFile, '@ ', Date);
          PrevDate := Date
        end;
      Amount := GetColumnReal (Line, 'EingegUmsatz');
      if GetColumnString (Line, 'EingegWaehr') = 'DEM' then
        Amount := Amount / 1.95583;
      if GetColumnString (Line, 'SollHaben') = 'S' then
        Amount := -Amount;
      Account := GetColumnString (Line, 'KtoNr');
      AntiAccount := GetColumnString (Line, 'Gegenkonto');
      VAT := GetColumnString (Line, 'BUFeld');
      if (VAT = '8') or (AntiAccount = '3300') then
        begin
          Command := 'gross ';
          VAT := ' VAT 0.07'
        end
      else if (VAT = '9') or (AntiAccount = '3400') then
        begin
          Command := 'gross ';
          VAT := ''
        end
      else if (Account = '8400') or (Account = '8401') then
        begin
          Command := 'antigross ';
          VAT := ''
        end
      else if (AntiAccount = '8400') or (AntiAccount = '8401') then
        begin
          Command := 'antigross ';
          TempStr := AntiAccount;
          AntiAccount := Account;
          Account := TempStr;
          Amount := - Amount;
          VAT := ''
        end
      else
        begin
          Command := 'noVAT ';
          VAT := ''
        end;
      Write (OutFile, Command, Amount : 0 : 2, ' from ', Account,
               ' for ', AntiAccount);
      Comment := GetColumnString (Line, 'BuchText');
      if Comment <> '' then
        Write (OutFile, ' "', Comment, '"');
      WriteLn (OutFile, VAT)
    end
end;

begin
  Init;
  AnalyseFirstLine;
  ProcessFile;
  Done
end.
